/*
 * indentation.cu
 *
 *  Created on: Apr 9, 2010
 *      Author: zhmurov
 */
#include "gsop.cuh"
#include "indentation.cuh"
#include "../Updaters/output_manager.cuh"
#include "indentation_kernel.cu"

int showTipMica;

void addTipMicaParticles();

/*
 * Add potential computation and update functions to lists,
 * create timer.
 */

void createIndentationPotential(){
	if(indentationOn == 1 || getYesNoParameter(INDENTATION_ON_STRING, 0, 1) == 1){
		indentationOn = 1;
		sprintf(indentationPotential.name, "Indentation");
		indentationPotential.compute = &computeIndentation;
		indentationPotential.computeEnergy = &computeIndentationEnergy;
		potentials[potentialsCount] = &indentationPotential;
		if(gsop.deviceProp.major == 2){
			cudaFuncSetCacheConfig(indentation_kernel, cudaFuncCachePreferL1);
		}
		potentialsCount++;
		sprintf(indentationUpdater.name, "Indentation");
		indentationUpdater.update = &updateTipPosition;
		indentationUpdater.frequency = nav;
		updaters[updatersCount] = &indentationUpdater;
		updatersCount++;
		initIndentation();
	}
}

/*
 * Initialize potential:
 * Tip and surface are both represented by a repulsive LJ-potential.
 * Tip is a sphere, connected to cantilever base by a spring.
 * Motion of a tip is described by Langevin equation but not integrated every step,
 * since tip is much heavier then particles.
 */
void initIndentation(){
	printf("Initializing indentation protocol...\n");
	// Read poteintial from configuration file
	getVectorParameter(INDENTATION_CHIP_POSITION_STRING, &indentation.chipCoord.x, &indentation.chipCoord.y, &indentation.chipCoord.z, 0.0, 0.0, 0.0, 0);
	indentation.chipCoord0 = indentation.chipCoord;
	getVectorParameter(INDENTATION_TIP_POSITION_STRING, &indentation.tipCoord.x, &indentation.tipCoord.y, &indentation.tipCoord.z,
			indentation.chipCoord.x, indentation.chipCoord.y, indentation.chipCoord.z, 1);
	getVectorParameter(INDENTATION_DIRECTION_STRING, &indentation.direction.x, &indentation.direction.y, &indentation.direction.z, 0.0, 0.0, 0.0, 0);
	getVectorParameter(INDENTATION_SURFACE_R0_STRING, &indentation.micaR0.x, &indentation.micaR0.y, &indentation.micaR0.z, 0.0, 0.0, 0.0, 0);
	indentation.micaR = indentation.micaR0;
	getVectorParameter(INDENTATION_SURFACE_N_STRING, &indentation.micaN.x, &indentation.micaN.y, &indentation.micaN.z, 0.0, 0.0, 0.0, 0);
	indentation.tipRadius = getFloatParameter(INDENTATION_TIP_RADIUS_STRING, 0.0, 0);
	indentation.cantileverKs = getFloatParameter(INDENTATION_KS_STRING, 0.0, 0);
	indentation.V = getFloatParameter(INDENTATION_DELTAX_STRING, 0.0, 0);
	indentation.dx = 0;
	indentation.moveSurface = getYesNoParameter(INDENTATION_MOVE_SURFACE, 0, 1);
	indentation.a6 = pow(getFloatParameter(INDENTATION_SIGMA_STRING, 1.0, 1), 6.0);
	indentation.el = getFloatParameter(INDENTATION_EL_STRING, 1.0, 1);
	indentation.tipZeta = getFloatParameter(INDENTATION_TIP_ZETA_STRING, 5000.0, 1);
	char indentation_filename[100];
	getMaskedParameter(indentation_filename, INDENTATION_OUTPUT_FILENAME_STRING, DEFAULT_INDENTATION_FILENAME, 1);

	printf("Initial cantilever base coordinates: (%5.2f, %5.2f, %5.2f).\n", indentation.chipCoord.x, indentation.chipCoord.y, indentation.chipCoord.z);

	// Normilize vectors
	float norm = sqrt(indentation.direction.x*indentation.direction.x + indentation.direction.y*indentation.direction.y + indentation.direction.z*indentation.direction.z);
	indentation.direction.x /= norm;
	indentation.direction.y /= norm;
	indentation.direction.z /= norm;
	norm = sqrt(indentation.micaN.x*indentation.micaN.x + indentation.micaN.y*indentation.micaN.y + indentation.micaN.z*indentation.micaN.z);
	indentation.micaN.x /= norm;
	indentation.micaN.y /= norm;
	indentation.micaN.z /= norm;

	// Allocatae memory, copy data to the device
	indentation.h_tipForces = (float4*)calloc(gsop.aminoCount, sizeof(float4));
	cudaMalloc((void**)&indentation.d_tipForces, gsop.aminoCount*sizeof(float4));
	cudaMemcpy(indentation.d_tipForces, indentation.h_tipForces, gsop.aminoCount*sizeof(float4), cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(c_indentation, &indentation, sizeof(Indentation), 0, cudaMemcpyHostToDevice);
	indentation.out = fopen(indentation_filename, "w");
	indentation.retractionStep = getLongIntegerParameter(INDENTATION_RETRACTION_STEP_STRING, -1, 1);
	checkCUDAError();

	// Add dummy particles to show surface and cantilever in dcd
	showTipMica = getYesNoParameter(INDENTATION_SHOW_TIP_SURFACE_STRING, 0, 1);
	if(showTipMica){
		sprintf(additionalAminosUpdater.name, "AdditionalAminos");
		additionalAminosUpdater.update = &updateAdditionalAminos;
		additionalAminosUpdater.frequency = getIntegerParameter(OUTPUT_FREQUENCY_STRING, DEFAULT_OUTPUT_FREQUENCY, 1);
		updaters[updatersCount] = &additionalAminosUpdater;
		updatersCount++;
		addTipMicaParticles();
	} else {
		sop.additionalAminosCount = 0;
	}
	indentation.outputFreq = getIntegerParameter(INDENTATION_OUTPUT_FREQ_STRING, DEFAULT_INDENTATION_OUTPUT_FREQ, 1);
}

/*
 * Create dummy particles, showing grid as a surface and a couple beads for cantalever
 */
void addTipMicaParticles(){
	int indentationMicaSize = getIntegerParameter(INDENTATION_MICA_SIZE_STRING, 51, 1);
	float indentationMicaStep = getFloatParameter(INDENTATION_MICA_STEP_STRING, 1.4, 1);
	sop.additionalAminosCount = 2 + indentationMicaSize*indentationMicaSize;
	sop.additionalAminos = (Atom*)calloc(sop.additionalAminosCount, sizeof(Atom));
	sop.additionalAminos[0].id = sop.aminoCount;
	strcpy(sop.additionalAminos[0].resName, "TIP");
	sop.additionalAminos[0].chain = 'T';
	sop.additionalAminos[0].x = indentation.chipCoord.x;
	sop.additionalAminos[0].y = indentation.chipCoord.y;
	sop.additionalAminos[0].z = indentation.chipCoord.z;
	sop.additionalAminos[0].resid = 1;
	sop.additionalAminos[1].id = sop.aminoCount + 1;
	strcpy(sop.additionalAminos[1].resName, "TIP");
	sop.additionalAminos[1].chain = 'T';
	sop.additionalAminos[1].x = indentation.tipCoord.x;
	sop.additionalAminos[1].y = indentation.tipCoord.y;
	sop.additionalAminos[1].z = indentation.tipCoord.z;
	sop.additionalAminos[1].resid = 2;
	float3 r1, r2, n, r0, r;
	n = indentation.micaN;
	r0 = indentation.micaR0;
	if(n.x == 0){
		r1.x = 1;
		r1.y = 0;
		r1.z = 0;
	} else if(n.y == 0){
		r1.x = 0;
		r1.y = 1;
		r1.z = 0;
	} else if(n.z == 0){
		r1.x = 0;
		r1.y = 0;
		r1.z = 1;
	} else {
		r1.x = 0;
		r1.y = 1;
		r1.z = - (n.y/n.z)*r1.y;
	}
	float norm = sqrt(r1.x*r1.x + r1.y*r1.y + r1.z*r1.z);
	r1.x /= norm;
	r1.y /= norm;
	r1.z /= norm;
	r2.x = n.y*r1.z - n.z*r1.y;
	r2.y = n.z*r1.x - n.x*r1.z;
	r2.z = n.x*r1.y - n.y*r1.x;
	int i, j;
	float displ = indentationMicaStep;
	for(i = 0; i < indentationMicaSize; i++){
		for(j = 0; j < indentationMicaSize; j++){
			int i1 = i-indentationMicaSize/2;
			int i2 = j-indentationMicaSize/2;
			r.x = r0.x + i1*displ*r1.x + i2*displ*r2.x;
			r.y = r0.y + i1*displ*r1.y + i2*displ*r2.y;
			r.z = r0.z + i1*displ*r1.z + i2*displ*r2.z;
			sop.additionalAminos[2+i*indentationMicaSize+j].id = sop.aminoCount + 2 + i*indentationMicaSize+j;
			strcpy(sop.additionalAminos[2+i*indentationMicaSize+j].resName, "MIC");
			sop.additionalAminos[2+i*indentationMicaSize+j].chain = 'M';
			sop.additionalAminos[2+i*indentationMicaSize+j].x = r.x;
			sop.additionalAminos[2+i*indentationMicaSize+j].y = r.y;
			sop.additionalAminos[2+i*indentationMicaSize+j].z = r.z;
		}
	}
	indentation.chipCoordAv.x = 0.0f;
	indentation.chipCoordAv.y = 0.0f;
	indentation.chipCoordAv.z = 0.0f;
	indentation.tipCoordAv.x = 0.0f;
	indentation.tipCoordAv.y = 0.0f;
	indentation.tipCoordAv.z = 0.0f;
	indentation.fav.x = 0.0f;
	indentation.fav.y = 0.0f;
	indentation.fav.z = 0.0f;
	indentation.fav.w = 0.0f;
	char tempFilename[100];
	getMaskedParameter(tempFilename, INDENTATION_VMD_CONNECT_SCRIPT, "connect_mica.vmd", 1);
	FILE* micaConnectFile = fopen(tempFilename, "w");
	printf("Dumping VMD script to connect surface elements into '%s'.\n", tempFilename);
	indentation.micaSize = indentationMicaSize;
	for(i = 0; i < indentationMicaSize; i++){
		for(j = 0; j < indentationMicaSize; j++){
			int i1 = 2+i*indentationMicaSize+j + gsop.aminoCount;
			int i2;
			fprintf(micaConnectFile, "set sel [atomselect top \"index %d\"]\n", i1);
			fprintf(micaConnectFile, "set bonds {");
			if(i != indentationMicaSize - 1){
				i2 = 2+(i+1)*indentationMicaSize+j + gsop.aminoCount;
				fprintf(micaConnectFile, "%d", i2);
			}
			if(j != indentationMicaSize - 1){
				i2 = 2+i*indentationMicaSize+j+1 + gsop.aminoCount;
				fprintf(micaConnectFile, " %d", i2);
			}
			fprintf(micaConnectFile, "}\n", i2);
			fprintf(micaConnectFile, "$sel setbonds [list $bonds]\n$sel delete\n");

		}
	}
	fclose(micaConnectFile);
}

/*
 * Start/stop timer, execute computations on GPU
 */
inline void computeIndentation(){
	indentation_kernel<<<gsop.aminoCount/BLOCK_SIZE + 1, BLOCK_SIZE>>>();
}

/*
 * Stub for energy evaluation
 */
inline void computeIndentationEnergy(){
	// TODO (if needed)
}


/*
 * Move cantilever
 */
inline void updateTipPosition(){
	// Change moving direction to switch to retraction
	if(step == indentation.retractionStep){
		printf("Starting tip retraction at step %ld.\n", indentation.retractionStep);
		/*indentation.direction.x = -indentation.direction.x;
		indentation.direction.y = -indentation.direction.y;
		indentation.direction.z = -indentation.direction.z;*/
		indentation.V = -indentation.V;
	}
	if(step % indentationUpdater.frequency == 0){
		// Copy forces acting onto tip to host memory
		cudaMemcpy(indentation.h_tipForces, indentation.d_tipForces, gsop.aminoCount*sizeof(float4), cudaMemcpyDeviceToHost);
		int i;
		float4 f = make_float4(0.0, 0.0, 0.0, 0.0);
		// Sum up all force and set 0 value to forces
		for(i = 0; i < gsop.aminoCount; i++){
			f.x += indentation.h_tipForces[i].x;
			f.y += indentation.h_tipForces[i].y;
			f.z += indentation.h_tipForces[i].z;
			indentation.h_tipForces[i].x = 0.0f;
			indentation.h_tipForces[i].y = 0.0f;
			indentation.h_tipForces[i].z = 0.0f;
		}
		cudaMemcpy(indentation.d_tipForces, indentation.h_tipForces, gsop.aminoCount*sizeof(float4), cudaMemcpyHostToDevice);
		// Compute average force and its absolute value
		f.x /= (float)indentationUpdater.frequency;
		f.y /= (float)indentationUpdater.frequency;
		f.z /= (float)indentationUpdater.frequency;
		f.w = sqrtf(f.x*f.x + f.y*f.y + f.z*f.z);
		indentation.tipForce = f;
		// Move chip or surface
		indentation.dx += indentation.V;//(step/nav)*indentation.chipV;
		if(indentation.moveSurface == 1){
			indentation.micaR.x = indentation.micaR0.x + indentation.direction.x*indentation.dx;
			indentation.micaR.y = indentation.micaR0.y + indentation.direction.y*indentation.dx;
			indentation.micaR.z = indentation.micaR0.z + indentation.direction.z*indentation.dx;
		} else {
			indentation.chipCoord.x = indentation.chipCoord0.x + indentation.direction.x*indentation.dx;
			indentation.chipCoord.y = indentation.chipCoord0.y + indentation.direction.y*indentation.dx;
			indentation.chipCoord.z = indentation.chipCoord0.z + indentation.direction.z*indentation.dx;
		}
		float4 dr;
		dr.x = indentation.chipCoord.x - indentation.tipCoord.x;
		dr.y = indentation.chipCoord.y - indentation.tipCoord.y;
		dr.z = indentation.chipCoord.z - indentation.tipCoord.z;
		dr.w = sqrtf(dr.x*dr.x + dr.y*dr.y + dr.z*dr.z);
		// Integrating equation of motion for cantilever tip (Langevin equation)
		float4 df;
		df.x = -f.x + dr.x*indentation.cantileverKs;
		df.y = -f.y + dr.y*indentation.cantileverKs;
		df.z = -f.z + dr.z*indentation.cantileverKs;
		df.w = sqrtf(df.x*df.x + df.y*df.y + df.z*df.z);

		float mult = ((integrator->h*((float)nav))/(indentation.tipZeta));

		indentation.tipCoord.x += mult*df.x;//indentationChipCoordX;// + dx*indentation.direction.x;
		indentation.tipCoord.y += mult*df.y;//indentationChipCoordY;// + dx*indentation.direction.y;
		indentation.tipCoord.z += mult*df.z;//indentationChipCoordZ;// + dx*indentation.direction.z;
		// Moving cantilever dummy beads for representation purposes

		indentation.chipCoordAv.x += indentation.chipCoord.x;
		indentation.chipCoordAv.y += indentation.chipCoord.y;
		indentation.chipCoordAv.z += indentation.chipCoord.z;
		indentation.tipCoordAv.x += indentation.tipCoord.x;
		indentation.tipCoordAv.y += indentation.tipCoord.y;
		indentation.tipCoordAv.z += indentation.tipCoord.z;
		indentation.fav.x += f.x;
		indentation.fav.y += f.y;
		indentation.fav.z += f.z;
		indentation.fav.w += f.w;
		indentation.kDeltaXAv += dr.w*indentation.cantileverKs;
		// Output the resulting data on the screen and into the file
		if(step % indentation.outputFreq == 0){
			int factor = indentation.outputFreq/indentationUpdater.frequency;
			indentation.chipCoordAv.x /= factor;
			indentation.chipCoordAv.y /= factor;
			indentation.chipCoordAv.z /= factor;
			indentation.tipCoordAv.x /= factor;
			indentation.tipCoordAv.y /= factor;
			indentation.tipCoordAv.z /= factor;
			indentation.fav.x /= factor;
			indentation.fav.y /= factor;
			indentation.fav.z /= factor;
			indentation.fav.w /= factor;
			indentation.kDeltaXAv /= factor;
			float fmod = sqrtf(indentation.fav.x*indentation.fav.x + indentation.fav.y*indentation.fav.y + indentation.fav.z*indentation.fav.z);
			float dxav = indentation.dx - 0.5*factor*indentation.V;
			printf("Indentation: %ld\t%5.2f\t%5.2f\t%5.2f\n", step, dxav, fmod, indentation.kDeltaXAv);
			fprintf(indentation.out, "%ld\t%8.5f\t"
					"%8.5f\t%8.5f\t%8.5f\t%8.5f\t%8.5f\t%8.5f\t%8.5f\t"
					"%8.5f\t%8.5f\t%8.5f\t"
					"%8.5f\t%8.5f\t%8.5f\n",
							step, indentation.dx,
							indentation.fav.w, fmod, dr.w*indentation.cantileverKs, indentation.kDeltaXAv, indentation.fav.x, indentation.fav.y, indentation.fav.z,
							indentation.tipCoord.x, indentation.tipCoord.y, indentation.tipCoord.z,
							indentation.chipCoord.x, indentation.chipCoord.y, indentation.chipCoord.z);
			fflush(indentation.out);
			indentation.chipCoordAv.x = 0.0f;
			indentation.chipCoordAv.y = 0.0f;
			indentation.chipCoordAv.z = 0.0f;
			indentation.tipCoordAv.x = 0.0f;
			indentation.tipCoordAv.y = 0.0f;
			indentation.tipCoordAv.z = 0.0f;
			indentation.fav.x = 0.0f;
			indentation.fav.y = 0.0f;
			indentation.fav.z = 0.0f;
			indentation.fav.w = 0.0f;
			indentation.kDeltaXAv = 0.0f;
		}
		// Update device data
		cudaMemcpyToSymbol(c_indentation, &indentation, sizeof(Indentation), 0, cudaMemcpyHostToDevice);
	}
}

void updateAdditionalAminos(){
	if(step % additionalAminosUpdater.frequency == 0){
		if(showTipMica){
			int i, j;
			sop.additionalAminos[1].x = indentation.tipCoord.x;
			sop.additionalAminos[1].y = indentation.tipCoord.y;
			sop.additionalAminos[1].z = indentation.tipCoord.z;
			if(indentation.moveSurface == 1){
				for(i = 0; i < indentation.micaSize; i++){
					for(j = 0; j < indentation.micaSize; j++){
						float mult = indentation.V*((float)additionalAminosUpdater.frequency)/((float)indentationUpdater.frequency);
						sop.additionalAminos[2+i*indentation.micaSize+j].x += mult*indentation.direction.x;
						sop.additionalAminos[2+i*indentation.micaSize+j].y += mult*indentation.direction.y;
						sop.additionalAminos[2+i*indentation.micaSize+j].z += mult*indentation.direction.z;
					}
				}
			} else {
				sop.additionalAminos[0].x = indentation.chipCoord.x;
				sop.additionalAminos[0].y = indentation.chipCoord.y;
				sop.additionalAminos[0].z = indentation.chipCoord.z;
			}
		}
	}
}
